//
// $Id: amoeba.cpp 2285 2010-05-08 10:28:53Z magnus $
//
// Project: Amoeba Program
// Authors: Magnus Leksell <nfk03mll@student.hig.se>, 
//          Wojciech Komorowski <nmd04wki@student.hig.se>
//

#include <iostream>
#include <sstream>
#include <fstream>
#include <typeinfo>
#include <cln/number.h>
#include <cln/complex.h>
#include <cln/float.h>
#include <cln/real.h>
#include <cln/lfloat.h>
#include <cln/rational.h>
#include <cln/integer.h>
#include <cln/univpoly_complex.h>
#include <cln/float_io.h>
#include <cln/real_io.h>
#include <cln/complex_io.h>

#include "amoeba.h"
#include "number.h"
#include "complex.h"
#include "square.h"
#include "circle.h"

using namespace std;
using namespace cln;

void AmoebaList::push_back(double x, double y) {
  list<AmoebaListDataType>::push_back(AmoebaListDataType(x, y));
}

Amoeba::Amoeba() {
}

Amoeba::~Amoeba() {
}

void Amoeba::setDimensions(int width, int height, double minLnW, double maxLnW,
                           double minLnZ, double maxLnZ) {
  imageWidth_ = width;
  imageHeight_ = height;
  minLnW_ = minLnW;
  maxLnW_ = maxLnW;
  minLnZ_ = minLnZ;
  maxLnZ_ = maxLnZ;
}

void Amoeba::setPrecision(int precision) {
  precision_ = precision;
}

void Amoeba::setMaxError(const cl_R& maxError) {
  maxError_ = maxError;
}

void Amoeba::setDataFileName(const string& dataFileName) {
  dataFileName_ = dataFileName;
}

void Amoeba::compute(const CPolynomial& p, const AbstractInterval* interval,
                     sem_t* mutex) {
  cl_inhibit_floating_point_underflow = false;

  // The list of (ln|z|, ln|w|) points that make up the amoeba
  AmoebaList amoebaList;

  // Determine the type of interval
  if (typeid(*interval) == typeid(Square))
  
    // Iterate through the rectangular region in the complex plane
    for (Number y=imagpart(interval->from()); y <= imagpart(interval->to());
         y = y + interval->step()) {
      for (Number x=realpart(interval->from()); x <= realpart(interval->to());
           x = x + interval->step()) {
        
        // This variable holds the constant which will substitute one of the
        // variables in p(z, w) = 0.
        cl_N subs = complex(x, y);
        
        // subs cannot be 0 since ln|0| is undefined
        if (zerop(subs))
          continue;
        
        // Now substitute each of the variables in p(z, w) with the
        // subs constant and thus create two univariate polynomials
        cl_UP_N p_z = createP_Z(p, subs);
        cl_UP_N p_w = createP_W(p, subs);

        // Take the natural logarithm of the absolute value of the
        // substituted variable
        double lnAbsSubs = double_approx(ln(abs(subs)));

        // Now find the roots of p_z(w) = 0 and p_w(z) = 0      
        list<cl_N> pz_roots =
          solver_.getRoots(p_z, float_format(precision_), maxError_);
        list<cl_N> pw_roots =
          solver_.getRoots(p_w, float_format(precision_), maxError_);

        list<cl_N>::const_iterator pzIter = pz_roots.begin();
        list<cl_N>::const_iterator pwIter = pw_roots.begin();

        // For each root of p_w(z) = 0...
        for (; pzIter != pz_roots.end(); ++pzIter) {
        
          // If it is 0, ignore it, since ln|0| is undefined
          if (zerop(*pzIter))
            continue;
            
          // Get the natural logarithm of the absolute value of the root
          double lnAbsRoot = double_approx(ln(abs(*pzIter)));
          
          // If p(z, w) = 0 but ln|z| or ln|w| falls outside the
          // user defined limits, then those values should be ignored...
          if (lnAbsSubs < minLnW_ || lnAbsSubs > maxLnW_)
            continue;
          if (lnAbsRoot < minLnZ_ || lnAbsRoot > maxLnZ_)
            continue;
          
          // Otherwise, we have a new (ln|z|, ln|w|) point of the amoeba!
          amoebaList.push_back(lnAbsRoot, lnAbsSubs);
        }
        // Now do the same for each root of p_z(w) = 0
        for (; pwIter != pw_roots.end(); ++pwIter) {
          if (zerop(*pwIter))
            continue;
          double lnAbsRoot = double_approx(ln(abs(*pwIter)));
          if (lnAbsSubs < minLnZ_ || lnAbsSubs > maxLnZ_)
            continue;
          if (lnAbsRoot < minLnW_ || lnAbsRoot > maxLnW_)
            continue;
          amoebaList.push_back(lnAbsSubs, lnAbsRoot);
        }
      }
    }
    
    // The interval is of type Circle
    else {
      // Iterate through all the circles
      for (Number radius = interval->minRadius()
          ; radius <= interval->maxRadius()
          ; radius = radius + interval->step()) {

        // Get the amount of substitute variables on each circle  
        int points = interval->points();

        cl_F Pi = pi(float_format(precision_));
        cl_F random = random_F(Pi);  // 0 <= random < Pi

        // For each substitute variable
        for (int i = 0; i < points; i++) {
        
          // Create the substitute!
          // the +random is used so that the points get more evenly
          // distributed around the circumference of each circle
          cl_N subs = radius*cis(((2*i*Pi)/points)+random);
          
          // Create the univariate polynomials by substituting one of
          // the variables in p(z, w)
          cl_UP_N p_z = createP_Z(p, subs);
          cl_UP_N p_w = createP_W(p, subs);
        
          // Find the roots of p_z(w) = 0 and p_w(z) = 0
          list<cl_N> pz_roots =
            solver_.getRoots(p_z, float_format(precision_), maxError_);
          list<cl_N> pw_roots =
            solver_.getRoots(p_w, float_format(precision_), maxError_);
          list<cl_N>::const_iterator pzIter = pz_roots.begin();
          list<cl_N>::const_iterator pwIter = pw_roots.begin();

          // Take the natural logarithm of the absolute value of the
          // substituted variable
          double lnAbsSubs = double_approx(ln(abs(subs)));

          // For each root of p_w(z) = 0...
          for (; pzIter != pz_roots.end(); ++pzIter) {
          
            // If it is 0, ignore it, since ln|0| is undefined
            if (zerop(*pzIter))
              continue;
            
            // Get the natural logarithm of the absolute value of the root
            double lnAbsRoot = double_approx(ln(abs(*pzIter)));

            // If p(z, w) = 0 but ln|z| or ln|w| falls outside the
            // user defined limits, then those values should be ignored...
            if (lnAbsSubs < minLnW_ || lnAbsSubs > maxLnW_)
              continue;
            if (lnAbsRoot < minLnZ_ || lnAbsRoot > maxLnZ_)
              continue;
            
            //Otherwise we have a new (ln|z|, ln|w|) point of the amoeba!
            amoebaList.push_back(lnAbsRoot, lnAbsSubs);
          }
          // Now do the same for each root of p_z(w) = 0
          for (; pwIter != pw_roots.end(); ++pwIter) {
            if (zerop(*pwIter)) continue;
            double lnAbsRoot = double_approx(ln(abs(*pwIter)));
            if (lnAbsSubs < minLnZ_ || lnAbsSubs > maxLnZ_)
              continue;
            if (lnAbsRoot < minLnW_ || lnAbsRoot > maxLnW_)
              continue;
            amoebaList.push_back(lnAbsSubs, lnAbsRoot);
          }
        }
      }
    }

  // mutex lock
  //cout << "Waiting..." << endl;
  if (sem_wait(mutex) != 0) {
    perror("Mutex wait error");
    exit(1);
  }
  //cout << "  writing... ";
  // Save to file
  ofstream os(dataFileName_.c_str(),
              ios_base::out | ios_base::binary | ios_base::app);
  AmoebaList::const_iterator iter = amoebaList.begin();
  for (; iter != amoebaList.end(); ++iter) {
    double x = (*iter).first;
    double y = (*iter).second;
    os.write((char*)&x, sizeof(double));
    os.write((char*)&y, sizeof(double));
  }
  os.close();
  //cout << "done." << endl;
  // mutex unlock
  if (sem_post(mutex) != 0) {
    perror("Mutex post error");
    exit(1);
  }
}

//
// This method maps the list of (ln|z|, ln|w|) points to an image with
// integer pixel coordinates
//
Points Amoeba::mapToImage(bool wantsAxes) throw(std::runtime_error) {
  Points points;
  AmoebaList amoebaList;
  
  // these are used to determine the minimum and maximum values of
  // all the ln|z|'s and ln|w|'s in the list, so that the image could
  // be scaled apropriately
  double minX, maxX, minY, maxY;

  minX = maxX = minY = maxY = 0;

  // Read points from data file
  ifstream is(dataFileName_.c_str(), ios_base::binary);

  if (!is)
    throw std::runtime_error(string("Could not open " + dataFileName_).c_str());

  while (is.good()) {
    double x;
    double y;
    is.read((char*)&x, sizeof(double));
    is.read((char*)&y, sizeof(double));
    if (x < minX)
      minX = x;
    if (x > maxX)
      maxX = x;
    if (y < minY)
      minY = y;
    if (y > maxY)
      maxY = y;
    amoebaList.push_back(x, y);
  }
  is.close();

  double dx = (maxX - minX)/(double)imageWidth_;
  double dy = (minY - maxY)/(double)imageHeight_;

  /* minX + (pixel_x)*dx = ln|z|
   * maxY + (pizel_z)*dy = ln|w|
   * pixel_x = (ln|z| - minX)/dx
   * pixel_y = (ln|w| - maxY)/dy
   */ 

  /**** DEBUG: make coord axes        ****/
  /**** shall we make this an option?  ***/
  if (wantsAxes) {
    double origo_x = -minX/dx;
    double origo_y = -maxY/dy;
    int foox = (int)origo_x;
    int barx = (int)origo_y;

    for (int i = 0; i < imageHeight_; i++) {
      Key key(foox, i);
      points[key] = DEFAULT_COLOR;
    }

    for (int i = 0; i < imageWidth_; i++) {
      Key key(i, barx);
      points[key] = DEFAULT_COLOR;
    }
  }
  /***********************************/

  // MOVE and adjust the points to fit an image structure.
  while (!amoebaList.empty()) {
    AmoebaListDataType xy = amoebaList.front();
    int Xpix = (int)((xy.first - minX)/dx);
    int Ypix = (int)((xy.second - maxY)/dy);
    Key key(Xpix, Ypix);
    points[key] = DEFAULT_COLOR;
    amoebaList.pop_front();
  }

  return points;
}

//
// Private method createP_Z
// Creates a polynomial p(z) from the given polynomial p(z, w)
// by substituting w in p(z, w) with a constant
//
cl_UP_N Amoeba::createP_Z(const CPolynomial& p, const cl_N& w) {
  cl_univpoly_complex_ring R = find_univpoly_ring(cl_C_ring);
  cl_UP_N p_z = R->zero();
  p_z.finalize();

  CPolynomial::const_iterator it;
  for (it = p.begin(); it != p.end(); ++it) {
    int z_exp = (*it).first.first;
    int w_exp = (*it).first.second;
    cl_UP_N tmp = R->create(z_exp);
    tmp.set_coeff(z_exp, expt(w, cl_I(w_exp))*(*it).second);
    tmp.finalize();
    p_z = p_z + tmp;
  }
  return p_z;
}

//
// Private method createP_W
// Creates a polynomial p(w) from the given polynomial p(z, w)
// by substituting z in p(z, w) with a constant
//
cl_UP_N Amoeba::createP_W(const CPolynomial& p, const cl_N& z) {
  cl_univpoly_complex_ring R = find_univpoly_ring(cl_C_ring);
  cl_UP_N p_w = R->zero();
  p_w.finalize();

  CPolynomial::const_iterator it;
  for (it = p.begin(); it != p.end(); ++it) {
    int z_exp = (*it).first.first;
    int w_exp = (*it).first.second;
    cl_UP_N tmp = R->create(w_exp);
    tmp.set_coeff(w_exp, expt(z, cl_I(z_exp))*(*it).second);
    tmp.finalize();
    p_w = p_w + tmp;
  }
  return p_w;
}
