//
// The Amoeba Generator is a program that computes and visualizes
// amoebas for some complex-valued bivariate expressions.
//
// Copyright (C) 2008 Magnus Leksell and Wojciech Komorowski.
//
// http://amoeba.sajberspejs.com
// Email: amoeba@sajberspejs.com
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#ifndef AMOEBA_H_
#define AMOEBA_H_

#include <cln/univpoly_complex.h>
#include <list>
#include <stdexcept>
#include <semaphore.h>
#include "rootfinders/CPSolver.h"
#include "interval.h"
#include "point.h"
#include "cpolynomial.h"

using namespace cln;

// The type of the values in the list of (ln|z|, ln|w|) points that make up
// the amoeba
typedef pair<double, double> AmoebaListDataType;

// The list itself
class AmoebaList : public list<AmoebaListDataType> {
public:
  void push_back(double x, double y);
};

class Amoeba {
  int imageWidth_;    // width of the image in pixels
  int imageHeight_;   // height of the image in pixels
  int precision_;     // number of bits in the mantissa part of each arbitrary
                      // precision number that is going to be used
  cl_R maxError_;

  // the limits on the values of (ln|z|, ln|w|) that make up the amoeba
  double minLnW_, maxLnW_, minLnZ_, maxLnZ_;

  // this stores the list of (ln|z|, ln|w|) points so that it can be reused
  // in consecutive runs of the program
  string dataFileName_;
  
  // the polynomial root finder
  CPSolver solver_;
public:
  Amoeba();
  ~Amoeba();
  
  // sets the image dimensions and limits on the values of ln|z| and ln|w|
  void setDimensions(int width, int height, double minLnW, double maxLnW,
                     double minLnZ, double maxLnZ);
                     
  // sets the amount of mantissa bits for arbitrary precision floating point
  // variables
  void setPrecision(int precision);
  
  void setMaxError(const cl_R& maxError);
  
  // sets a name for the data file
  void setDataFileName(const string& dataFileName);
  
  /*
    This method computes the amoeba of the bivariate polynomial p(z, w)
    by replacing one of the variables by a constant to get an univariate
    polynomial, say p_z(w). The roots of p_z(w) = 0 are then located, and
    the pair (ln|z|, ln|root|) is inserted into a list, for all the roots
    of p_z(w). The same procedure is applied by replacing the other variable
    by a constant to get p_w(z) = 0 and adding (ln|root, ln|w|) into the list.
    The set of constant values which replace one of the variables in p(z, w)
    is taken from an AbstractInterval representing a Square or a Circle.
    If the AbstractInterval represents a Square, the interval is a rectangular
    region in the complex plane where the set of constants are evenly
    distributed. Otherwise it is a disk centered at the origin,
    modelled as a set of circles centered at the origin, with increasing radius,
    and the set of constants is evenly distributed along the circumference of
    each circle.
  */
  void compute(const CPolynomial& p, const AbstractInterval* interval,
               sem_t* mutex);

  // maps the list of (ln|z|, ln|w|) points to pixels
  Points mapToImage(bool wantsAxes = false) throw(std::runtime_error);
  //void mapToImage(const char* fileName) throw(std::runtime_error);

private:

  // these methods substitute one of the variables in the bivariate polynomial
  // p(z, w) to create an univariate polynomial p_z(w) or p_w(z)
  cl_UP_N createP_Z(const CPolynomial& p, const cl_N& w);
  cl_UP_N createP_W(const CPolynomial& p, const cl_N& z);
};

#endif /* AMOEBA_H_ */
