//
// The Amoeba Generator is a program that computes and visualizes
// amoebas for some complex-valued bivariate expressions.
//
// Copyright (C) 2008 Magnus Leksell and Wojciech Komorowski.
//
// http://amoeba.sajberspejs.com
// Email: amoeba@sajberspejs.com
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include <iostream>
#include <sstream>
#include <stdexcept>
#include <cln/cln.h>
#include <sys/types.h>
#include <regex.h>
#include "conffile.h"
#include "square.h"
#include "circle.h"

using namespace cln;

//
// Constructor
//
ConfFile::ConfFile(bool verbose) {
  onDescription_ = false;
  onPowerSeries_ = false;
  firstTime_ = true;
  verbose_ = verbose;
}

//
// Destructor
//
ConfFile::~ConfFile() {
  // Deallocate intervals
  while (!intervals_.empty()) {
    AbstractInterval* i = intervals_.front();
    delete i;
    intervals_.pop_front();
  }
}

//
// Private method createLF
//
// Creates a cl_LF with correct precision.
//
cl_LF ConfFile::createLF(const string& x) {
  ostringstream os;
  //os << x + "L+0_" << precision_;
  os << x << "_" << precision_;

  return cl_LF(os.str().c_str());
  /*
  cl_LF lf(os.str().c_str());
  cout << "DEBUG: createLF(" << x << ") -> "
       << os.str() << " -> " << lf << endl;
  return lf;
  */
}

//
// Private method createComplexLF
//
// Creates a cl_N with realpart and imagpart as cl_LF.
//
cl_N ConfFile::createComplexLF(const string& x) {
  // Read a complex
  istringstream is(x);
  cl_N dummy;
  is >> dummy;

  //float_format_t prec = float_format(precision_);

  // Create cl_LF of the realpart
  ostringstream os;
  os << realpart(dummy);
  cl_LF a = createLF(os.str());
  //cl_F a = cl_float(realpart(dummy), prec);

  // Create cl_LF of the imagpart
  os.str("");
  os << imagpart(dummy);
  cl_LF b = createLF(os.str());
  //cl_F b = cl_float(imagpart(dummy), prec);

  return complex(a, b);
}

//
// Private method createTaylor
// This method creates the Taylor expansion for some functions so that
// functions of the type f(z, w) = z + cos(w) can be approximated by
// bivariate polynomials p(z, w) = z + 1 - w^2/2! + ...
//
void ConfFile::createTaylor(const cl_N& coeff, const string& fun,
                            const string& var, unsigned int terms) {
  if (var != "z" && var != "w")
    throw runtime_error("Unknown variable. Check spelling in XML file");

  if (fun == "sin")
    taylorexpSin(coeff, var, terms);
  else if (fun == "cos")
    taylorexpCos(coeff, var, terms);
  else if (fun == "exp")
    taylorexpExp(coeff, var, terms);
  else
    throw runtime_error("Unknown function. Check spelling in XML file");
}

//
// Creates a Taylor expansion for the sine function
// sin z = z - z^3/3! + z^5/5! - ...
//
void ConfFile::taylorexpSin(const cl_N& coeff, const string& var,
                            unsigned int terms) {
  bool negative = false;
  bool z_var = (var == "z");

  for (unsigned int i = 0; i < terms; i++) {
    unsigned int exponent = 2*i+1;
    cl_I fact = factorial(exponent);
    if (z_var) {
      cl_N tmp = poly_[Key(exponent, 0)];
      if (negative)
        tmp = tmp - coeff/fact;
      else
        tmp = tmp + coeff/fact;
      poly_[Key(exponent, 0)] = tmp;
    }
    else {
      cl_N tmp = poly_[Key(0, exponent)];
      if (negative)
        tmp = tmp - coeff/fact;
      else
        tmp = tmp + coeff/fact;
      poly_[Key(0, exponent)] = tmp;
    }
    negative = !negative;
  }
}

//
// Creates a Taylor expansion for the cosine function
// cos z = 1 - z^2/2! + z^4/4! - ...
//
void ConfFile::taylorexpCos(const cl_N& coeff, const string& var,
                            unsigned int terms) {
  bool negative = false;
  bool z_var = (var == "z");

  for (unsigned int i = 0; i < terms; i++) {
    unsigned int exponent = 2*i;
    cl_I fact = factorial(exponent);
    if (z_var) {
      cl_N tmp = poly_[Key(exponent, 0)];
      if (negative)
        tmp = tmp - coeff/fact;
      else
        tmp = tmp + coeff/fact;
      poly_[Key(exponent, 0)] = tmp;
    }
    else {
      cl_N tmp = poly_[Key(0, exponent)];
      if (negative)
        tmp = tmp - coeff/fact;
      else
        tmp = tmp + coeff/fact;
      poly_[Key(0, exponent)] = tmp;
    }
    negative = !negative;
  }
}

//
// Creates a Taylor expansion for the exponential function
// exp z = 1 + x + x^2/2! + x^3/3! + ...
//
void ConfFile::taylorexpExp(const cl_N& coeff, const string& var,
                            unsigned int terms) {
  bool z_var = (var == "z");

  for (unsigned int i = 0; i < terms; i++) {
    cl_I fact = factorial(i);
    if (z_var) {
      cl_N tmp = poly_[Key(i, 0)];
      tmp = tmp + coeff/fact;
      poly_[Key(i, 0)] = tmp;
    }
    else {
      cl_N tmp = poly_[Key(0, i)];
      tmp = tmp + coeff/fact;
      poly_[Key(0, i)] = tmp;
    }
  }
}

//
// Private method validate
//
bool ConfFile::validate(const string& exp) {
  if (exp == "j" || exp == "-j")
    return true;

  // This accepts patterns like (n, m integer):
  // 0, j, -j, j+n, j-n, mj, -mj, mj+n, mj-n, -mj+n, -mj-n, m, -m
  // Or shortly, [-][m][j][+-][n]
  string expr = "^\\([0]\\)\\|\\([-]\\{0,1\\}\\([1-9][0-9]*\\)\\{0,1\\}";
         expr += "\\)\\{0,1\\}[j]\\{0,1\\}";
         expr += "\\([+-][1-9][0-9]*\\)\\{0,1\\}$";
  regex_t preg;

  // Compile the expression
  if (regcomp(&preg, expr.c_str(), REG_NOSUB) != 0) {
    // This should never happend!
    cerr << "regcomp error!" << endl;
    return false;
  }

  // Validate the exponent
  int status = regexec(&preg, exp.c_str(), (size_t)0, NULL, 0);
  regfree(&preg);

  return status == 0;
}

//
// Private method eval
//
// Evaluates the exponent string to a integer.
//
// Precondition: That the string exp is a valid
// expression according to the validate method.
//
int ConfFile::eval(const string& exp, unsigned int j) {
  int m = 1;
  int n = 0;
  int xj = 1;
  string str = exp;
  int jindex = exp.find_first_of('j');

  if (str == "j")
    xj = j;
  else if (str == "-j")
    xj = -j;
  else if (jindex == 1 && exp[0] == '-') { // "-j..."
    str.replace(0, 2, 1, ' ');             // Remove the "-j" substring
    xj = -j;
    istringstream is(str);
    is >> n;                               // Read one integer
  }
  else if (jindex == 0) {                  // "j..."
    str.replace(jindex, 1, 1, ' ');        // Remove the 'j' character
    xj = j;
    istringstream is(str);
    is >> n;                               // Read one integer
  }
  else if (jindex > 0) {                   // "...j..."
    str.replace(jindex, 1, 1, ' ');        // Remove the 'j' character
    xj = j;
    istringstream is(str);
    is >> m;                               // Get the first integer
    if (is.good())                         // If there is a second integer:
      is >> n;                             //   get it
  }
  else {                                   // No 'j' in exp
    istringstream is(str);
    is >> m;                               // Get the first integer
    if (is.good())                         // If there is a second integer:
      is >> n;                             //   get it
  }

  int value = m * xj + n;

  if (verbose_) {
    cout << "    j=" << j << ", eval(" << exp << "): " << m << "*" << xj
         << (n<0?"":"+") << n << "=" << value << endl;
  }

  return value;
}

//
// Private method createPowerSeries
//
void ConfFile::createPowerSeries(const cl_N& coeff, const string& var,
                                 const string& cexp, const string& vexp,
                                 const string& jcoeff, const string& fun) {
  if (var != "z" && var != "w")
    throw runtime_error("Unknown variable. Check spelling in XML file");

  bool z_var = (var == "z");

  for (unsigned int j = powerSeriesRange_.first
    ; j <= powerSeriesRange_.second; j++) {

    int zexp = 0;
    int wexp = 0;

    if (z_var)
      zexp = eval(vexp, j);
    else
      wexp = eval(vexp, j);

    // Validate exponents
    if (zexp < 0)
      throw out_of_range("z-exponent below 0");

    if (wexp < 0)
      throw out_of_range("w-exponent below 0");

    int cj = eval(cexp, j);
    cl_I jcoeffValue = 1;

    if (jcoeff != "") {
      jcoeffValue = eval(jcoeff, j);
      if (fun != "") {
        if (fun == "fac")
          jcoeffValue = factorial(cl_I_to_uint(jcoeffValue));
        else if (fun == "abs")
          jcoeffValue = abs(jcoeffValue);
        else
          throw runtime_error("Unknown function. Check spelling in XML file");
      }
    }

    cl_N tmp;

    /*if (z_var)
      tmp = zPoly_[zexp];
    else
      tmp = wPoly_[wexp];*/

    tmp = poly_[Key(zexp, wexp)];

    if (tmp == complex(0, 0))
      tmp = complex(1, 0);

    // DEBUG
    /*cout << "tmp=" << tmp << ", "
         << "coeff=" << coeff << ", "
         << "jcoeffValue=" << jcoeffValue << ", "
	 << "cj=" << cj << endl;*/

    // multiply with current coeff
    tmp = tmp * expt(jcoeffValue, cj) * expt(coeff, cj);

    /*if (z_var)
      zPoly_[zexp] = tmp;
    else
      wPoly_[wexp] = tmp;*/

    poly_[Key(zexp, wexp)] = tmp;

    if (verbose_) {
      cout << "    added: (" << tmp
           << ")" << "z^" << zexp << "w^" << wexp << endl;
    }
  }
}

//
// Private method addSeriesToPolyString
//
void ConfFile::addSeriesToPolyString(const cl_N& coeff, const string& var,
                                     const string& cexp, const string& vexp,
                                     const string& jcoeff, const string& fun) {
  ostringstream os;

  // TeXify the power series
  if (firstTime_) {
    os << "\\sum_{j=" << powerSeriesRange_.first << "}^{"
       << powerSeriesRange_.second << "}";
    firstTime_ = false;
  }

  if (realpart(coeff) != 1 || imagpart(coeff) != 0)
    os << "(" << coeff << ")^{" << cexp << "}";

  if (jcoeff != "") {
    if (fun != "") {
      if (fun == "fac")
        os << "[(" << jcoeff << ")!]^{" << cexp << "}";
      else
        os << "[\\" << fun << "(" << jcoeff << ")]^{" << cexp << "}";
    }
    else
      os << "(" << jcoeff << ")^{" << cexp << "}";
  }

  os << var << "^{" << vexp << "}";

  polyString_ += os.str();
}

//
// Private method addToPolyString
//
void ConfFile::addToPolyString(const cl_N& coeff, unsigned int zexp,
                               unsigned int wexp) {
  ostringstream os;

  if (!firstTime_ && realpart(coeff) >= 0)
    os << "+";

  if (zexp == 0 && wexp == 0)
    os << coeff;
  else if (realpart(coeff) != 1 || imagpart(coeff) != 0)
    if (imagpart(coeff) == 0)
      os << coeff;
    else
      os << "(" << coeff << ")";

  if (zexp == 1)
    os << "z";
  else if (zexp != 0)
    os << "z^" << zexp;

  if (wexp == 1)
    os << "w";
  else if (wexp != 0)
    os << "w^" << wexp;

  firstTime_ = false;

  polyString_ += os.str();
}

//
// Private method addToPolyString
//
void ConfFile::addToPolyString(const cl_N& coeff, const string& fun,
                               const string& var) {
  ostringstream os;

  if (!firstTime_ && realpart(coeff) >= 0)
    os << "+";

  if (realpart(coeff) != 1 || imagpart(coeff) != 0)
    os << coeff;

  // TeXify the function
  os << "\\" << fun << "(" << var << ")";

  firstTime_ = false;
  polyString_ += os.str();
}

//
// Private method setAmoeba
//
void ConfFile::setAmoeba(const AttributeList& properties) {
  AttributeList::const_iterator iter = properties.begin();
  for (; iter != properties.end(); ++iter) {
    if (iter->name == "name")
      amoebaName_ = iter->value;
    else
      cout << "Warning: Unknown XML entity: " << iter->name << endl;
  }
}

//
// Private method setSettings
//
void ConfFile::setSettings(const AttributeList& properties) {
/*
  AttributeList::const_iterator iter = properties.begin();
  for (; iter != properties.end(); ++iter) {
    //cout << "  Attribute " << iter->name << " = " << iter->value << endl;
    if (iter->name == "precision")
      precision_ = atoi(iter->value.c_str());
    else if (iter->name == "iterations")
      iterations_ = atoi(iter->value.c_str());
    else if (iter->name == "rooterror")
      ; // TODO: rootError_ = ?
  }
*/
}

//
// Private method setImage
//
void ConfFile::setImage(const AttributeList& properties) {
  AttributeList::const_iterator iter = properties.begin();
  for (; iter != properties.end(); ++iter) {
    //cout << "  Attribute " << iter->name << " = " << iter->value << endl;
    if (iter->name == "filename")
      imageFileName_ = iter->value;
    else if (iter->name == "width")
      imageWidth_ = atoi(iter->value.c_str());
    else if (iter->name == "height")
      imageHeight_ = atoi(iter->value.c_str());
    else if (iter->name == "axes")
      wantsAxes_ = iter->value == "yes";
    else
      cout << "Warning: Unknown XML entity: " << iter->name << endl;
  }
}

//
// Private method setLatex
//
void ConfFile::setLatex(const AttributeList& properties) {
  AttributeList::const_iterator iter = properties.begin();
  for (; iter != properties.end(); ++iter) {
    //cout << "  Attribute " << iter->name << " = " << iter->value << endl;
    if (iter->name == "filename")
      latexFileName_ = iter->value;
    else
      cout << "Warning: Unknown XML entity: " << iter->name << endl;
  }
}

//
// Private method setData
//
void ConfFile::setData(const AttributeList& properties) {
  AttributeList::const_iterator iter = properties.begin();
  for (; iter != properties.end(); ++iter) {
    //cout << "  Attribute " << iter->name << " = " << iter->value << endl;
    if (iter->name == "filename")
      dataFileName_ = iter->value;
    else
      cout << "Warning: Unknown XML entity: " << iter->name << endl;
  }
}

//
// Private method setNPolytope
//
void ConfFile::setNPolytope(const AttributeList& properties) {
  AttributeList::const_iterator iter = properties.begin();
  for (; iter != properties.end(); ++iter) {
    //cout << "  Attribute " << iter->name << " = " << iter->value << endl;
    if (iter->name == "filename")
      newtonPolytopeFileName_ = iter->value;
    else
      cout << "Warning: Unknown XML entity: " << iter->name << endl;
  }
}

//
// Private method setPolynomial
//
void ConfFile::setPolynomial(const AttributeList& properties) {
  if (verbose_)
    cout << "Parsing polynomial..." << endl;

  AttributeList::const_iterator iter = properties.begin();
  for (; iter != properties.end(); ++iter) {
    //cout << "  Attribute " << iter->name << " = " << iter->value << endl;
  }
}

//
// Private method setPowerSeries
//
void ConfFile::setPowerSeries(const AttributeList& properties) {
  unsigned int from = 0;
  unsigned int to = 0;
  cl_N coeff = complex(1, 0);
  string var = "";

  if (verbose_)
    cout << "Parsing power series, ";

  AttributeList::const_iterator iter = properties.begin();
  for (; iter != properties.end(); ++iter) {
    if (iter->name == "from")
      from = atoi(iter->value.c_str());
    else if (iter->name == "to")
      to = atoi(iter->value.c_str());
    else
      cout << "Warning: Unknown XML entity: " << iter->name << endl;
  }

  if (verbose_)
    cout << "j=" << from << ".." << to << endl;

  // Verify some values
  if (from >= to)
    throw runtime_error("to-value not larger than from-value");

  if (to > 1024)
    throw out_of_range("to-value larger than 1024");

  onPowerSeries_ = true;
  powerSeriesRange_.first = from;
  powerSeriesRange_.second = to;
}

//
// Private method setTerm
//
void ConfFile::setTerm(const AttributeList& properties) {
  cl_N coeff = complex(1, 0);
  unsigned int zexp = 0;
  unsigned int wexp = 0;
  bool gotFunction = false;
  string fun = "";
  string var = "";
  string cexp = "j"; // coeff exponent
  string vexp = "j"; // var exponent
  unsigned int terms = 10; // default
  string jcoeff = "";

  AttributeList::const_iterator iter = properties.begin();
  for (; iter != properties.end(); ++iter) {
    if (iter->name == "coeff") {
      istringstream is(iter->value.c_str());
      is >> coeff;
    }
    else if (iter->name == "jcoeff") {
      jcoeff = iter->value;
    }
    else if (iter->name == "zexp")
      zexp = atoi(iter->value.c_str());
    else if (iter->name == "wexp")
      wexp = atoi(iter->value.c_str());
    else if (iter->name == "function") {
      fun = iter->value;
      gotFunction = true;
    }
    else if (iter->name == "var")
      var = iter->value;
    else if (iter->name == "terms")
      terms = atoi(iter->value.c_str());
    else if (iter->name == "cexp")
      cexp = iter->value;
    else if (iter->name == "vexp")
      vexp = iter->value;
    else
      cout << "Warning: Unknown XML entity: " << iter->name << endl;
  }

  // Print some info
  if (verbose_)
    if (onPowerSeries_) {
      cout << "  adding term: (" << coeff << ")^{" << cexp << "}";
      if (jcoeff != "") {
        if (gotFunction)
          cout << "[" << fun << "(" << jcoeff << ")]^{" << cexp << "}";
        else
          cout << "(" << jcoeff << ")^{" << cexp << "}";
      }
      if (var != "")
        cout << var << "^{" << vexp << "}";
      cout << endl;
    }
    else if (gotFunction)
      cout << "  adding function: (" << coeff << ")"
           << fun << "(" << var << "), expansion terms = " << terms << endl;
    else
      cout << "  adding term: (" << coeff << ")z^"
           << zexp << "w^" << wexp << endl;

  // Verify some values
  if (zexp < 0 || zexp > 1024)
    throw out_of_range("z exponent out of range 0..1024");

  if (wexp < 0 || wexp > 1024)
    throw out_of_range("w exponent out of range 0..1024");

  if (terms < 0 || terms > 1024)
    throw out_of_range("terms out of range 0..1024");

  if (!validate(cexp))
    throw runtime_error("coeff-exponent not a valid j-expression");

  if (!validate(vexp))
    throw runtime_error("var-exponent not a valid j-expression");

  if (jcoeff != "" && !validate(jcoeff))
    throw runtime_error("jcoeff not a valid j-expression");

  // Add to the polynomial
  if (onPowerSeries_) {
    addSeriesToPolyString(coeff, var, cexp, vexp, jcoeff, fun);
    createPowerSeries(coeff, var, cexp, vexp, jcoeff, fun);
  }
  else if (gotFunction) {
    addToPolyString(coeff, fun, var);
    createTaylor(coeff, fun, var, terms);
  }
  else {
    addToPolyString(coeff, zexp, wexp);
    poly_[Key(zexp, wexp)] = coeff;
  }
}

//
// Private method setInterval
//
void ConfFile::setInterval(const AttributeList& properties) {
  Glib::ustring type;
  Complex from = complex(0, 0);
  Complex to = complex(0, 0);
  Number minRadius = 0.0;
  Number maxRadius = 0.0;
  Number step = 0.0;
  unsigned int points = 0;

  AttributeList::const_iterator iter = properties.begin();
  for (; iter != properties.end(); ++iter) {
    if (iter->name == "type")
      type = iter->value;
    else if (iter->name == "minradius")
      minRadius = createLF(iter->value);
    else if (iter->name == "maxradius")
      maxRadius = createLF(iter->value);
    else if (iter->name == "step")
      step = createLF(iter->value);
    else if (iter->name == "points")
      points = atoi(iter->value.c_str());
    else if (iter->name == "from") {
      from = createComplexLF(iter->value);
    }
    else if (iter->name == "to") {
      to = createComplexLF(iter->value);
    }
    else
      cout << "Warning: Unknown XML entity: " << iter->name << endl;
  }

  if (type == "circle")
    intervals_.push_back(new Circle(minRadius, maxRadius, step, points));
  else if (type == "square")
    intervals_.push_back(new Square(from, to, step));
  else
    throw runtime_error("Unknown interval type");
}

//
// Private method setMax
//
void ConfFile::setMax(const AttributeList& properties) {
  AttributeList::const_iterator iter = properties.begin();
  for (; iter != properties.end(); ++iter) {
    //cout << "  Attribute " << iter->name << " = " << iter->value << endl;
    if (iter->name == "precision")
      precision_ = atoi(iter->value.c_str());
    else if (iter->name == "error")
      maxError_ = cl_R(iter->value.c_str());
    else if (iter->name == "processes")
      maxProcesses_ = atoi(iter->value.c_str());
    else if (iter->name == "lnZ")
      maxLnZ_ = atof(iter->value.c_str());
    else if (iter->name == "lnW")
      maxLnW_ = atof(iter->value.c_str());
    else
      cout << "Warning: Unknown XML entity: " << iter->name << endl;
  }

  // Validate some data
  if (precision_ < 15 || precision_ > 1024)
    throw out_of_range("precision out of range 15..1024");

  if (maxError_ < cl_R("1e-30") || maxError_ > cl_R("1e-2"))
    throw out_of_range("max error out of range 1e-30..1e-2");

  if (maxProcesses_ < 1 || maxProcesses_ > 1024)
    throw out_of_range("max processes out of range 1..1024");
}

//
// Private method setMin
//
void ConfFile::setMin(const AttributeList& properties) {
  AttributeList::const_iterator iter = properties.begin();
  for (; iter != properties.end(); ++iter) {
    if (iter->name == "lnZ")
      minLnZ_ = atof(iter->value.c_str());
    else if (iter->name == "lnW")
      minLnW_ = atof(iter->value.c_str());
    else
      cout << "Warning: Unknown XML entity: " << iter->name << endl;
  }
}

void ConfFile::on_start_document() {
  //cout << "on_start_document()" << endl;
}

void ConfFile::on_end_document() {
  //cout << "on_end_document()" << endl;
}

void ConfFile::on_start_element(const Glib::ustring& name,
                                const AttributeList& attributes) {
  // Check node name (case sensitive) and call appropriate method.
  if (name == "amoeba")
    setAmoeba(attributes);
  else if (name == "description")
    onDescription_ = true;
  else if (name == "settings")
    setSettings(attributes);
  else if (name == "image")
    setImage(attributes);
  else if (name == "latex")
    setLatex(attributes);
  else if (name == "data")
    setData(attributes);
  else if (name == "npolytope")
    setNPolytope(attributes);
  else if (name == "polynomial")
    setPolynomial(attributes);
  else if (name == "powerseries")
    setPowerSeries(attributes);
  else if (name == "term")
    setTerm(attributes);
  else if (name == "interval")
    setInterval(attributes);
  else if (name == "max")
    setMax(attributes);
  else if (name == "min")
    setMin(attributes);
  else
    cerr << "Warning: Unknown XML entity: " << name << endl;
}

void ConfFile::on_end_element(const Glib::ustring& name) {
  //cout << "on_end_element()" << endl;
}

void ConfFile::on_characters(const Glib::ustring& text) {
  //cout << "on_characters(): " << text << endl;
  if (onDescription_) {
    description_ = text;
    onDescription_ = false;
  }
}

void ConfFile::on_comment(const Glib::ustring& text) {
  //cout << "on_comment(): " << text << endl;
}

void ConfFile::on_warning(const Glib::ustring& text) {
  //cout << "on_warning(): " << text << endl;
}

void ConfFile::on_error(const Glib::ustring& text) {
  //cout << "on_error(): " << text << endl;
}

void ConfFile::on_fatal_error(const Glib::ustring& text) {
  //cout << "on_fatal_error(): " << text << endl;
}

/*
const CPolynomial& ConfFile::getPolynomial() {
  if (onPowerSeries_) {
    // Construct polynomial from the two series
    map<int, cl_N>::const_iterator wIter = wPoly_.begin();
    for (; wIter != wPoly_.end(); ++wIter) {
      map<int, cl_N>::const_iterator zIter = zPoly_.begin();
      for (; zIter != zPoly_.end(); ++zIter) {
        int zexp = zIter->first;
        int wexp = wIter->first;
	cl_N prod = zIter->second * wIter->second;

        poly_[Key(zexp, wexp)] = prod;

	if (verbose_)
          cout << "Added: " << prod << "z^" << zexp << "w^" << wexp << endl;
      }
    }
  }
  return poly_;
}
*/

void ConfFile::parseFile(const string& filename) {
  set_substitute_entities(true); // what is this and why?
  parse_file(filename);
}

