//
// $Id: main.cpp 2285 2010-05-08 10:28:53Z magnus $
//
// Project: Amoeba Program
// Authors: Magnus Leksell <nfk03mll@student.hig.se>, 
//          Wojciech Komorowski <nmd04wki@student.hig.se>
//

#include <iostream>
#include <fstream>
#include <sstream>
#include <argp.h>
#include <stdexcept>
#include <semaphore.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/shm.h>
#include <sys/stat.h>
#include "environment.h"
#include "conffile.h"
#include "amoeba.h"
#include "image.h"
#include "listplotter.h"
#include "latex.h"
#include "npolytope.h"

using namespace std;

const char *argp_program_version = "Amoeba Generator 1.0";
const char *argp_program_bug_address = "<amoeba@sajberspejs.com>";
static char doc[] = "Amoeba Generator -- a program for plotting amoebas.";
static char args_doc[] = "";

static struct argp_option options[] = {
  {"plot", 'p', "LISTFILE", 0,
   "Plots an image from a predefined list of x, y values" },
  {"output", 'o', "IMAGEFILE", 0,
   "Name of the png file to plot - used in conjunction with --plot or -p" },
  {"width", 'w', "WIDTH", 0,
   "Width of the image to generate from the list" },
  {"height", 'h', "HEIGHT", 0,
   "Height of the image to generate from the list" },
  {"amoeba", 'a', "AMOEBA.xml", 0,
   "Plots an amoeba as defined in an XML file" },
  {"skip", 's', 0, 0,
   "Skip computing the amoeba, use data file from previous run - \
used in conjunction with -a" },
  {"verbose", 'v', 0, 0, "Verbose mode. Dump values read from XML file etc."},
  { 0 }
};

struct arguments {
  bool plot, amoeba, skip, verbose;
  char* list_file;
  char* image_file;
  int width, height;
  char* amoeba_file;
};

static error_t parse_opt (int key, char *arg, struct argp_state *state) {
  struct arguments *arguments = (struct arguments*) state->input;
  switch (key) {
    case 'a':
      arguments->amoeba = true;
      arguments->amoeba_file = arg;
      break;
    case 's':
      arguments->skip = true;
      break;
    case 'v':
      arguments->verbose = true;
      break;
    case 'p':
      arguments->plot = true;
      arguments->list_file = arg;
      break;
    case 'o':
      arguments->image_file = arg;
      break;
    case 'w':
      arguments->width = atoi(arg);
      break;
    case 'h':
      arguments->height = atoi(arg);
      break;
    case ARGP_KEY_ARG:
      //if (state->arg_num >= 2)
        //argp_usage (state);
      break;
    case ARGP_KEY_END:
      if (state->argc < 2)
        argp_usage (state);
      //if (state->arg_num < 2)
        //argp_usage (state);
      break;
    default:
      return ARGP_ERR_UNKNOWN;
  }
  return 0;
}

/* Our argp parser. */
static struct argp argp = {options, parse_opt, args_doc, doc};

int main(int argc, char** argv) {
  key_t shm_key;
  int shm_id;
  sem_t* mutex;
  struct arguments arguments;

  // Create shared memory
  shm_id = shmget(IPC_PRIVATE, sizeof(sem_t), IPC_CREAT | 0600);
  if (shm_id == -1) {
    perror("Could not get shared memory");
    return 1;
  }

  // Attach to shared memory
  mutex = (sem_t*)shmat(shm_id, (void *)0, 0);
  if ((long)mutex == -1) {
    perror("Could not attach to shared memory");
    return 1;
  }

  // Create mutex
  if (sem_init(mutex, 1, 1) != 0) {
    perror("Could not create mutex");
    return 1;
  }

  arguments.plot = false;
  arguments.amoeba = false;
  arguments.skip = false;
  arguments.verbose = false;
  arguments.list_file = strdup("-");
  arguments.image_file = strdup("from_list.png");
  arguments.width = 640;
  arguments.height = 480;
  arguments.amoeba_file = strdup("-");
  
  argp_parse(&argp, argc, argv, 0, 0, &arguments);

  if (arguments.plot)
    try {
      // Plot image and then quit
      ListPlotter::plotList(arguments.list_file, arguments.image_file,
                            arguments.width, arguments.height);
      return 0;
    }
    catch (const exception& e) {
      cerr << "Exception caught: " << e.what() << endl;
      return 1;
    }

  if (!arguments.amoeba) {
    cerr << "Could not understand arguments." << endl;
    return 1;
  }

  string infile = arguments.amoeba_file;
  Environment* env = new ConfFile(arguments.verbose);

  // Try parse the config file.
  try {
    env->parseFile(infile);
    if (arguments.verbose)
      cout << *env;
  }
  catch (const exception& e) {
    cerr << "Exception caught: " << e.what() << endl;
    return 1;
  }

  if (arguments.skip) // Skip computing the amoeba
    cout << "Skipping computing the amoeba." << endl
         << "Using data from " << env->getDataFileName() << endl;
  
  // Set some initial values.
  int imgWidth = env->getImageWidth();
  int imgHeight = env->getImageHeight();
  double minLnW = env->getMinLnW();
  double maxLnW = env->getMaxLnW();
  double minLnZ = env->getMinLnZ();
  double maxLnZ = env->getMaxLnZ();

  Amoeba amoeba;
  amoeba.setDataFileName(env->getDataFileName());
  amoeba.setDimensions(imgWidth, imgHeight, minLnW, maxLnW, minLnZ, maxLnZ);
  amoeba.setPrecision(env->getPrecision());
  amoeba.setMaxError(env->getMaxError());

  if (!arguments.skip) { // Calculate the amoeba
    // Remove the data file before computing, if it exists
    remove(env->getDataFileName().c_str());

    cout << "Saving data in " << env->getDataFileName() << endl
         << "Using " << env->getMaxProcesses()
         << " parallel processes." << endl;

    int intervalCount = 0; // Keep track of number of intervals processed
    int processCount = 0;  // Number of parallel processes currently running
    CPolynomial poly = env->getPolynomial();
    const list<AbstractInterval*> intervals = env->getIntervals();
    list<AbstractInterval*>::const_iterator iter = intervals.begin();

    while (iter != intervals.end()) { // For each interval
      pid_t child_pid;

      ++intervalCount;

      // Start a new process
      child_pid = fork();

      if (child_pid == 0) { // Child process
        cout << "Computing #" << intervalCount
             << "(" << intervals.size() << "): " << **iter << " " << endl;
        amoeba.compute(poly, *iter, mutex);
        if (arguments.verbose)
          cout << "...done with #" << intervalCount << endl;
        exit(0); // Finished, just exit child process
      }
      else if (child_pid == -1) {
        perror("Could not start new process");
        exit(1);
      }

      // Parent process
      ++processCount; // Increase number of processes running

      // If maximum number of processes running,
      // wait for any process to finish
      if (processCount == env->getMaxProcesses()) {
        wait(NULL);     // Waiting for a process to finish
        --processCount; // Decrease number of processes running
      }

      ++iter; // Next interval

    } // while intervals...

    // Wait for remaining childs to finish
    while (processCount > 0) {
      wait(NULL);
      --processCount;
    }
  } // !arguments.skip

  // Remove the data file when calculations is done
  //if (!arguments.skip)
    //remove(env->getDataFileName().c_str());

  // Detach and remove shared memory
  if (shmdt(mutex) == 0)
    shmctl(shm_id, IPC_RMID, NULL);

  try {
    // Map the amoeba to an image structure
    // and write the image to a file
    const char* imgfile = env->getImageFileName().c_str();
    Points points = amoeba.mapToImage(env->wantsAxes());
    ImageProducer::write(imgfile, points, env->getImageWidth(),
                         env->getImageHeight());
    cout << "Saved image in " << imgfile << endl;

    // Produce the Newton polytope image
    const char* npfile = env->getNewtonPolytopeFileName().c_str();
    NewtonPolytope::write(npfile, env->getPolynomial());

    // Produce a LaTeX file.
    LatexProducer::write(*env);
  }
  catch (const exception& e) {
    cerr << "Exception caught: " << e.what() << endl;
    return 1;
  }

  // Done!
  return 0;
}
