//
// The Amoeba Generator is a program that computes and visualizes
// amoebas for some complex-valued bivariate expressions.
//
// Copyright (C) 2008 Magnus Leksell and Wojciech Komorowski.
//
// http://amoeba.sajberspejs.com
// Email: amoeba@sajberspejs.com
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include <iostream>
#include <cmath>
#include <stdexcept>
#include <deque>
#include <gd.h>
#include "npolytope.h"
#include "point.h"

using namespace std;

void NewtonPolytope::write(const char* filename,
                           const CPolynomial& p) throw(runtime_error) {

  if (filename == "") // Empty string -- no file name given:
    return;           // silently skip this.

  int width;
  int height;
  gdImagePtr im;

  gdPoint points[p.size()];
  int index = 0;
  int maxX = 0, maxY = 0;
  CPolynomial::const_iterator iter = p.begin();
  for (; iter != p.end(); ++iter) {
    Key key = (*iter).first;
    unsigned int zexp = key.first;
    unsigned int wexp = key.second;
    points[index].x = zexp;
    points[index].y = wexp;
    if (zexp > maxX) maxX = zexp;
    if (wexp > maxY) maxY = wexp;
    ++index;
  }

  // if we can't draw a polytope because of too few
  // points, then silently skip it
  if (p.size() < 2)
    return;
  
  // find the pivot
  int pIndex = 0;   // pivot index
  for (int i = 0; i < p.size(); i++) {
    if (points[i].y < points[pIndex].y)
      pIndex = i;
    else if (points[i].y == points[pIndex].y && points[i].x < points[pIndex].x)
      pIndex = i;
  }

  // place the pivot at the beginning
  if (pIndex != 0) {
    gdPoint tmp = points[0];
    points[0] = points[pIndex];
    points[pIndex] = tmp;
    pIndex = 0;
  }
    
  // sort the rest by increasing angle relative to the pivot
  int tmpIndex;
  double tmpAngle;
  gdPoint tmpPoint;
  gdPoint pivot = points[0];
  int j;
  for (int i = 1; i < p.size(); i++) {
    tmpIndex = i;
    tmpPoint = points[i];
    tmpAngle = angle(pivot, points[i]);
    for (j = i; j < p.size(); j++) {
      if (angle(pivot, points[j]) <= tmpAngle) {
        tmpPoint = points[j];
        tmpAngle = angle(pivot, tmpPoint);
        tmpIndex = j;
      }
    }
    points[tmpIndex] = points[i];
    points[i] = tmpPoint;
  }

  // if several points have the same angle relative to the pivot,
  // delete them except the one with the furthest distance from
  // the pivot
  deque<gdPoint> ptsQue;
  for (int i = 0; i < p.size(); i++) {
    ptsQue.push_back(points[i]);
  }
  {
    gdPoint p0 = ptsQue[0];
    gdPoint p1, p2;
    for (int i = 1; i < ptsQue.size()-1; i++) {
      p1 = ptsQue[i];
      p2 = ptsQue[i+1];
      if (angle(p0, p1) == angle(p0, p2)) {
        if ((p0.x-p1.x)*(p0.x-p1.x)+(p0.y-p1.y)*(p0.y-p1.y) <
            (p0.x-p2.x)*(p0.x-p2.x)+(p0.y-p2.y)*(p0.y-p2.y)) {
          i--;
          ptsQue.erase(ptsQue.begin() + (i+1));
        }
        else {
          i--;
          ptsQue.erase(ptsQue.begin() + (i+2));
        }
      }
    }
  }

  // construct the convex hull
  deque<gdPoint> hullQue;
  hullQue.push_back(ptsQue[0]);
  hullQue.push_back(ptsQue[1]);
  gdPoint p1, p2;
  int i = 2;
  while (i < ptsQue.size()) {
    p1 = hullQue[hullQue.size()-2];
    p2 = hullQue[hullQue.size()-1];
    if (leftTurn(p1, p2, ptsQue[i])) {
      hullQue.push_back(ptsQue[i]);
      i++;
    }
    else {
      hullQue.pop_back();
    }
  }

  i = hullQue.size();
  int hullpts = i;
  gdPoint hull[i];
  for (i--; i > -1; i--) {
    hull[i] = hullQue.back();
    hullQue.pop_back();
  }

  // setup the image
  int scale = 35;
  int radius = 4;
  width  = scale*maxX + radius/2 + 3;
  height = scale*maxY + radius/2 + 3;
  im = gdImageCreate(width, height);
  int white = gdImageColorAllocate(im, 255, 255, 255); // Background color white
  int black = gdImageColorAllocate(im, 0, 0, 0);
  int grey  = gdImageColorAllocate(im, 153, 153, 153);
  int blue  = gdImageColorAllocate(im, 0, 0, 255);

  // map all points to the image
  for (int i = 0; i < p.size(); i++) {
    int zexp = points[i].x;
    int wexp = points[i].y;
    points[i].x = (zexp * scale) + radius/2;
    points[i].y = height - (wexp * scale) - radius/2 - 1;
  }
  for (int i = 0; i < hullpts; i++) {
    int zexp = hull[i].x;
    int wexp = hull[i].y;
    hull[i].x = (zexp * scale) + radius/2;
    hull[i].y = height - (wexp * scale) - radius/2 - 1;
  }
    
  // draw points together with the hull
  for (int i = 0; i < p.size(); i++) {
    int cx = points[i].x;
    int cy = points[i].y;
    gdImageFilledArc(im, cx, cy, radius, radius, 0, 360, grey, gdArc);
  }
  for (int i = 0; i < hullpts-1; i++) {
    gdImageLine(im, hull[i].x, hull[i].y, hull[i+1].x, hull[i+1].y, black);
  }
  gdImageLine(im, hull[hullpts-1].x, hull[hullpts-1].y, points[0].x, points[0].y, black);
  
  //gdImagePolygon(im, points, p.size(), black);

  FILE* pngout = fopen(filename, "wb");
  if (pngout == NULL)
    throw runtime_error("Could not save image to file!");

  gdImagePng(im, pngout);
  gdImageDestroy(im);

  cout << "Saved Newton Polytope in " << filename << endl;
}

double NewtonPolytope::angle(gdPoint p1, gdPoint p2)
{
  double pi = 3.14159265358979;
  int Vx = p2.x - p1.x;
  int Vy = p2.y - p1.y;
  double norm = sqrt(Vx*Vx + Vy*Vy);
  if (norm == 0.0) return norm;
  
  double cosangle = (double)Vx / norm;
  if (p1.y < p2.y) return acos(cosangle);
  else if (p1.y > p2.y) return 2*pi*acos(cosangle);
  else if (p1.x > p2.x) return pi;
  else return 0.0;
}

bool NewtonPolytope::leftTurn(gdPoint p1, gdPoint p2, gdPoint p3)
{
  return ((p2.x - p1.x)*(p3.y - p1.y) - (p2.y - p1.y)*(p3.x - p1.x)) >= 0;
}

