//
// $Id: CPSolver.cpp 2285 2010-05-08 10:28:53Z magnus $
//
// CPSolver.cpp
//
// Project: Amoeba Program
// Authors: Magnus Leksell      <nfk03mll@student.hig.se>,
//          Wojciech Komorowski <nmd04wki@student.hig.se>
//
// Description:
// An implementation of the Durand-Kerner method for simultaneously finding
// all the roots of polynomials of arbitrary degree with complex coefficients.
///////////////////////////////////////////////////////////////////////////////

#include <list>
#include <iostream>

#include <cln/number.h>
#include <cln/complex.h>
#include <cln/real.h>
#include <cln/float.h>
#include <cln/sfloat.h>
#include <cln/ffloat.h>
#include <cln/dfloat.h>
#include <cln/lfloat.h>
#include <cln/rational.h>
#include <cln/integer.h>
#include <cln/univpoly_complex.h>

#include <cln/complex_io.h>
#include <cln/lfloat_io.h>

#include "CPSolver.h"

using namespace std;
using namespace cln;

// The Durand-Kerner iterative method for polynomial root finding
list<cl_N> CPSolver::getRoots(const cl_UP_N &p, float_format_t precision,
                                                  const cl_R& maxError)
{
  list<cl_N> roots;             // where the roots will be stored
  sintL p_degree = degree(p);   // degree of the polynomial

  // solve explicitly for simple polynomials
  if (p_degree < 1) {
    cerr << "ERROR: rootfinder called with polynomial degree < 1" << endl;
    abort();
  }
  if (p_degree == 1) {
    cl_N root = -coeff(p, 0)/coeff(p, 1);
    roots.push_back(cl_N(root));
    return roots;
  }
  else if (p_degree == 2) {
    cl_N a = coeff(p, 2);
    cl_N b = coeff(p, 1);
    cl_N c = coeff(p, 0);
    cl_N discroot = sqrt(b*b-4*a*c);
    roots.push_back(cl_N((-b+discroot)/(2*a)));
    roots.push_back(cl_N((-b-discroot)/(2*a)));
    return roots;
  }

  /* set starting values by distributing them around a circle which is known
   * to contain all the roots of the polynomial - see (1), end of Section 6.
   */

  // consider all roots to be a cluster and take its center
  // to be the mean of all the roots, i.e. -C_(n-1)/(nC_n)
  //cl_N center = -(coeff(p, p_degree-1)/(p_degree*coeff(p, p_degree)));

/*
  // calculate the radius
  cl_RA one(1);
  cl_R radius = 1.0;
  for (int i = 1; i <= p_degree; i++) {
    cl_UP_N P = derive(p, p_degree-i);
    cl_R foo = abs(P(center)/coeff(p, p_degree));
    cl_F n = cl_float((int)p_degree, precision);
    cl_I fact = factorial(p_degree-i);
    n = n/fact;
    cl_N expr = n*foo;
    cl_RA blah(i);
    cl_N exponent = one/blah;
    cl_R tmp = realpart(expt(expr, exponent));
    if (tmp > radius)
      radius = tmp;
  }
*/

  // distribute n numbers around the circle...
  // {z_(1,..,n)} = {center + radius * e^(2ikPi/n), k=1...n-1}
  cl_N r_approx[p_degree];
  cl_F Pi = pi(precision);
  for (int i = 1; i < p_degree; i++)
    //r_approx[i] = center + radius * cis((2*i*Pi)/p_degree);
    r_approx[i] = cis((2*i*Pi)/p_degree);

  // now add the final point, so that it doesn't end up on the real axis
  //r_approx[0] = center + radius * cis(Pi/p_degree);
  r_approx[0] = cis(Pi/p_degree);

  cl_N r_prod[p_degree];

  // the leading coefficient of the polynomial must be 1 for
  // this method to converge...
  cl_univpoly_complex_ring R = find_univpoly_ring(cl_C_ring);
  cl_UP_N C_n = R->create(0);
  C_n.set_coeff(0, 1/coeff(p, p_degree));
  C_n.finalize();
  cl_UP_N P = C_n*p;
  
  for (unsigned int iter = 1; iter <= p_degree; iter++) {
    for (int i = 0; i < p_degree; i++) {
      cl_N prod = complex(1, 0);
      for (int j = 0; j < p_degree; j++) {
        if (j == i) continue;
        prod = prod * (r_approx[i] - r_approx[j]);
      }
      cl_N tmp = r_approx[i];
      r_approx[i] = tmp - (P(tmp)/prod);
    }
  }

  unsigned int iterations = p_degree;

  for (int i = 0; i < p_degree; i++) {
    cl_R error = abs(p(r_approx[i]));
    if (error > maxError) {
      bool convergence; 
      do {
        convergence = true;
        for (int i = 0; i < p_degree; i++) {
          cl_N prod = complex(1, 0);
          for (int j = 0; j < p_degree; j++) {
            if (j == i) continue;
            prod = prod * (r_approx[i] - r_approx[j]);
          }
          cl_N tmp = r_approx[i];
          r_approx[i] = tmp - (P(tmp)/prod);
          iterations++;
        }
        if (iterations >= 65535) {
          cerr << "Root finder failed after " << iterations << " iterations!" << endl;
          for (int i = 0; i < p_degree; i++)
            roots.push_back(cl_N(r_approx[i]));
          return roots;
        }
        for (int i = 0; i < p_degree; i++) {
          error = abs(p(r_approx[i]));
          if (error > maxError) {
            convergence = false;
            break;
          }
        }
      }
      while (!convergence);
      break;
    }
  }

  for (int i = 0; i < p_degree; i++)
    roots.push_back(cl_N(r_approx[i]));

  return roots;
}

inline cl_UP_N CPSolver::derive(const cl_UP_N &p, unsigned int n)
{
  cl_UP_N tmp = p;
  for (unsigned int i = 0; i < n; i++)
    tmp = deriv(tmp);
  return tmp;
}

